/* Just some base styles not needed for example to function */
*, html { font-family: Verdana, Arial, Helvetica, sans-serif; }

body, form, ul, li, p, h1, h2, h3, h4, h5
{
	margin: 0;
	padding: 0;
}
body { background-color: #606061; color: #ffffff; margin: 0; }
img { border: none; }
p
{
	font-size: 1em;
	margin: 0 0 1em 0;
}

html { font-size: 100%; /* IE hack */ }
body { font-size: 1em; /* Sets base font size to 16px */ }
table { font-size: 100%; /* IE hack */ }
input, select, textarea, th, td { font-size: 1em; }
	
/* CSS Tree menu styles */
ol.tree
{
	padding: 0 0 0 30px;
	width: 300px;
}
	li 
	{ 
		position: relative; 
		margin-left: -15px;
		list-style: none;
	}
	li.dot
	{
		margin-left: -1px !important;
	}
		li.dot a
		{
			background-repeat: no-repeat;
			background-position: 0 0;
			background-image: url(dot.png);
			color: #fff;
			padding-left: 21px;
			text-decoration: none;
			display: block;
		}
	li.file
	{
		margin-left: -1px !important;
	}
		li.file a
		{
			background-repeat: no-repeat;
			background-position: 0 0;
			background-image: url(document.png);
			color: #fff;
			padding-left: 21px;
			text-decoration: none;
			display: block;
		}
	li input
	{
		position: absolute;
		left: 0;
		margin-left: 0;
		opacity: 0;
		z-index: 2;
		cursor: pointer;
		height: 1em;
		width: 1em;
		top: 0;
	}
		li input + ol
		{
			background-repeat: no-repeat;
			background-position: 40px 0;
			background-image: url(toggle-small-expand.png);
			margin: -0.938em 0 0 -44px; /* 15px */
			height: 1em;
		}
		li input + ol > li { display: none; margin-left: -14px !important; padding-left: 1px; }
	li label
	{
		background-repeat: no-repeat;
		background-position: 15px 1px;
		background-image: url(folder-horizontal.png);
		cursor: pointer;
		display: block;
		padding-left: 37px;
	}

	li input:checked + ol
	{
		background-repeat: no-repeat;
		background-position: 40px 5px;
		background-image: url(toggle-small.png);
		margin: -1.25em 0 0 -44px; /* 20px */
		padding: 1.563em 0 0 80px;
		height: auto;
	}
		li input:checked + ol > li { display: block; margin: 0 0 0.125em;  /* 2px */}
		li input:checked + ol > li:last-child { margin: 0 0 0.063em; /* 1px */ }